// Nathan E. Moore
// April, 2, 2004

#pragma once
#include "parser.h"
#include "idlist.h"
#include "menuentry.h"
#include <list>
#include <vector>
#include <string>
#include <algorithm>
#include <functional>
#include <boost/bind.hpp>
#include <boost/mem_fn.hpp>
#include <boost/shared_ptr.hpp>

// in a idea world this would be ';', just like the default path separator on windows
// but I suppose that that would just make to much sense
static const wchar_t G_PATH_SEPARATOR = '|';

template <bool> class FolderParser;

template <bool> ParserPtr buildFolderParser(const std::wstring& txt, const wchar_t* paths);

namespace fp_impl
{
	// fwd declarations
	class MergeableInterface;
	class Path;

	// typedefs
	typedef std::list<MenuEntryPtr> EntryList;

	// functions
	void parseFolder(const Path& parsepath, MergeableInterface& mi);

	// abstract base classes
	class NOVTABLE MergeableInterface
	{
	public:
		virtual void addEntry(const MenuEntryPtr& item) = 0;
		virtual void addSubMenu(const std::wstring& txt, const std::wstring& path) = 0;
	};

	// classes
	class Path
	{
	private:
		void validate();
	public:
		// Path
		std::wstring m_path;
		// a '?' and '*' pattern;
		std::wstring m_file;

		Path(const std::wstring& directory, const std::wstring& pattern)
			:m_path(directory),
			m_file(pattern)
		{
			validate();
		}

		Path(const std::wstring& completepath)
			:m_path(completepath),
			m_file()
		{
			validate();
		}

		std::wstring getSearchPath() const
		{
			return std::wstring(L"\\\\?\\") + m_path + m_file;
		}
	};

	// Order is a template because both MenuEntryInterface and ParserInterface require a
	// getEntryText method and thus Order can be used to compare them.
	template <class lhs, class rhs = lhs>
	class Order : public std::binary_function<lhs, rhs, bool>
	{
	private:
		static inline bool nocase(wchar_t x, wchar_t y)
		{
			// remember towupper is not a localized function;
			return towupper(x) < towupper(y);
		}

	public:
		bool operator()(const lhs& left, const rhs& right) const
		{
			const std::wstring& lhsstr = left->getEntryText();
			const std::wstring& rhsstr = right->getEntryText();
			return std::lexicographical_compare(lhsstr.begin(), lhsstr.end(), rhsstr.begin(), rhsstr.end(), nocase);
		}
	};

	template <bool t_dynamic>
	class Mergeable : public MergeableInterface
	{
	private:
		typedef std::list<boost::shared_ptr<::FolderParser<t_dynamic> > > FPList;
		EntryList* m_elist;
		FPList m_newparsers;

	public:
		FPList m_plist;
		unsigned int m_subcount;
		unsigned int m_entrycount;

		Mergeable(EntryList* elist)
			:m_elist(elist),
			m_subcount(0),
			m_entrycount(0)
		{}

		void addEntry(const MenuEntryPtr& item)
		{
			m_elist->push_back(item);
			++m_entrycount;
		}

		void parse(const Path& path);
		void addSubMenu(const std::wstring& txt, const std::wstring& path);
	};
}	// fp_impl

template <bool t_dynamic>
class FolderParser : public ParserBase<t_dynamic>
{
private:
	typedef std::list<MenuSubPtr> SubList;
	typedef std::list<fp_impl::Path> PathList;
	PathList m_paths;
	fp_impl::EntryList m_entries;
	SubList m_subs;

	FolderParser(const std::wstring& txt)
		:ParserBase<t_dynamic>(txt)
	{}

public:
	void clearStorage()
	{
		m_entries.clear();
		m_subs.clear();
	}

	bool doContextMenu(HwndInterface& hi)
	{
		return util::popupContextMenu(m_paths.front().m_path, hi);
	}

	void addPath(const std::wstring& path)
	{
		m_paths.push_back(fp_impl::Path(path));
	}

	unsigned int parse(HmenuInterface& menu, unsigned int start);
	MenuSubPtr findSubmenu(HMENU hmenu) const;
	MenuEntryPtr findEntry(unsigned int id) const;

	static boost::shared_ptr<FolderParser<t_dynamic> > createFolderParser(const std::wstring& txt)
	{
		boost::shared_ptr<FolderParser<t_dynamic> > temp(new FolderParser(txt));
		return temp;
	}
};

template <bool t_dynamic>
void fp_impl::Mergeable<t_dynamic>::addSubMenu(const std::wstring& txt, const std::wstring& path)
{
	// find_if m_slist contains a parser with EntryText the same as txt
	FPList::iterator i = std::find_if(m_plist.begin(), m_plist.end(), boost::bind<bool>(std::equal_to<std::wstring>(), txt,
	                                  boost::bind<const std::wstring&>(boost::mem_fn<const std::wstring&, ParserInterface>(&ParserInterface::getEntryText), _1)));
	if (m_plist.end() != i)
		(*i)->addPath(path);
	else
	{
		boost::shared_ptr<FolderParser<t_dynamic> > fparser(FolderParser<t_dynamic>::createFolderParser(txt));
		fparser->addPath(path);
		m_newparsers.push_back(fparser);
		++m_subcount;
	}
}

template <bool t_dynamic>
void fp_impl::Mergeable<t_dynamic>::parse(const Path& path)
{
	parseFolder(path, *this);
	Order<boost::shared_ptr<FolderParser<t_dynamic> > > order;
	m_newparsers.sort(order);
	m_plist.merge(m_newparsers, order);
}


template <bool t_dynamic>
MenuEntryPtr FolderParser<t_dynamic>::findEntry(unsigned int id) const
{
	// in here
	fp_impl::EntryList::const_iterator iter = std::find_if(m_entries.begin(), m_entries.end(), EntrySearcher(id));
	if (iter != m_entries.end())
		return *iter;

	// in a submenu
	// TODO: get rid of the recursion
	for (SubList::const_iterator i = m_subs.begin(), end = m_subs.end(); i != end; ++i)
	{
		MenuEntryPtr item = (*i)->getEntry(id);
		if (item)
			return item;
	}

	// not in this menu hierarchy
	return MenuEntryPtr();
}

template <bool t_dynamic>
MenuSubPtr FolderParser<t_dynamic>::findSubmenu(HMENU hmenu) const
{
	// in here
	SubList::const_iterator iter = std::find_if(m_subs.begin(), m_subs.end(), SubSearcher(hmenu));
	if (iter != m_subs.end())
		return *iter;

	// in a submenu
	// TODO: get rid of the recursion
	for (SubList::const_iterator i = m_subs.begin(), end = m_subs.end(); i != end; ++i)
	{
		MenuSubPtr sub = (*i)->getSub(hmenu);
		if (sub)
			return sub;
	}

	// not in this menu hierarchy
	return MenuSubPtr();
}

template <bool t_dynamic>
unsigned int FolderParser<t_dynamic>::parse(HmenuInterface& menu, unsigned int start)
{
	fp_impl::Mergeable<t_dynamic> merger(&m_entries);
	// reference to reference problem forces the use of a explicit loop
	//std::for_each(m_paths.begin(), m_paths.end(), std::bind1st(std::mem_fun(&Mergeable<t_dynamic>::parse), &merger));
	for (PathList::const_iterator i = m_paths.begin(), end = m_paths.end(); i != end; ++i)
		merger.parse(*i);

	for (fp_impl::Mergeable<t_dynamic>::FPList::const_iterator i = merger.m_plist.begin(), end = merger.m_plist.end(); i != end; ++i, ++start)
	{
		MenuSubPtr sub = MenuSub::createMenuSub(*i, true);
		m_subs.push_back(sub);
		menu.insertSubmenu(sub, start);
	}

	m_entries.sort(fp_impl::Order<MenuEntryPtr>());
	std::for_each(m_entries.rbegin(), m_entries.rend(), boost::bind(HmenuInterface::insertEntry, &menu, _1, start, false));
	return merger.m_subcount + merger.m_entrycount;
}

template <bool t_dynamic>
ParserPtr buildFolderParser(const std::wstring& txt, const wchar_t* paths)
{
	boost::shared_ptr<FolderParser<t_dynamic> > parser(FolderParser<t_dynamic>::createFolderParser(txt));
	const wchar_t* start = paths;
	const wchar_t* end = paths + lstrlen(paths);
	const wchar_t* pos;
	while (end != (pos = std::find(start, end, G_PATH_SEPARATOR)))
	{
		parser->addPath(std::wstring(start, pos));
		start = pos + 1;
	}
	parser->addPath(std::wstring(start, end));
	return boost::static_pointer_cast<ParserPtr::element_type>(parser);
}
